#define _GNU_SOURCE

#include "print_error.h" // inline
#include "struct_CPInfo.h"
#include "struct_OPArg.h"
#include "write_log_hash_check.h" // inline

#include "md5.h"

#include <stdbool.h>
#include <stdlib.h>

#include <fcntl.h>
#include <malloc.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define BUF_NUM 2

static void *buf[BUF_NUM];
static int fd;
static int buffer_size_local;
static long psize;
static md5_state_t state;
static md5_byte_t digest[16];
static int read_size[BUF_NUM];
static off_t total;
static off_t static_to_size;
static int r_number;
static int a_number;
static _Bool lseek_err;
static _Bool read_err;
static _Bool read_finish;

/* 関数プロトタイプ */
void verify_hash(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd);
static void * read_func(void *dummy);
static void * append_func(void *dummy);

/*******************************************************************************
*******************************************************************************/
void verify_hash(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd)
{
	static _Bool get_memory = false;
	fd = 0;
	buffer_size_local = oparg->buffer_size / 2;
	read_size[0] = 0;
	read_size[1] = 0;
	total = 0;
	static_to_size = cpinfo->to_size;
	lseek_err = false;
	read_err = false;
	read_finish = false;

	_Bool open_flag = false;

	if(psize == 0)
	{
		psize = sysconf(_SC_PAGESIZE);
	}

	errno = 0;
	if((fd = open(cpinfo->to, O_NOATIME | O_NOFOLLOW | O_DIRECT | O_RDONLY)) == -1)
	{
		switch(errno)
		{
		case EPERM:
			if((fd = open(cpinfo->to, O_NOFOLLOW | O_DIRECT | O_RDONLY)) == -1)
			{
				print_error("open", __FILE__, __LINE__, cpdd);
			}
			else
			{
				open_flag = true;
			}
			break;

		default:
			print_error("open", __FILE__, __LINE__, cpdd);
			break;
		}
	}
	else
	{
		open_flag = true;
	}

	if(open_flag == true)
	{
		if((buf[0] == NULL) &&(buf[1] == NULL))
		{
			if((posix_memalign((void **)&buf[0], psize, buffer_size_local) != 0) || (posix_memalign((void **)&buf[1], psize, buffer_size_local) != 0))
			{
				fprintf(stderr, "ベリファイに失敗しました\n");
				fprintf(stderr, "%s\n", cpinfo->to);
				print_error("posix_memalign", __FILE__, __LINE__, cpdd);
				write_log_hash_check(cpinfo, cpinfo->verify_md5, "NULL", sdir, cpdd, false);
			}
			else
			{
				get_memory = true;
			}
		}

		switch(get_memory)
		{
		case true:
			md5_init(&state);

			r_number = 0;
			a_number = 1;
			pthread_t t1;
			pthread_t t2;

			pthread_create(&t1, NULL, read_func, (void *)&r_number);
			pthread_join(t1, NULL);

			for(;;)
			{
				r_number++;
				a_number--;
				pthread_create(&t1, NULL, read_func, (void *)&r_number);
				pthread_create(&t2, NULL, append_func, (void *)&a_number);

				pthread_join(t1, NULL);
				pthread_join(t2, NULL);

				if((read_finish == true) || (lseek_err == true) || (read_err == true))
				{
					break;
				}

				r_number--;
				a_number++;
				pthread_create(&t1, NULL, read_func, (void *)&r_number);
				pthread_create(&t2, NULL, append_func, (void *)&a_number);

				pthread_join(t1, NULL);
				pthread_join(t2, NULL);

				if((read_finish == true) || (lseek_err == true) || (read_err == true))
				{
					break;
				}
			}

			md5_finish(&state, digest);

			static char hex_output[(16 * 2) + 1];
			hex_output[0] = '\0';

			for(int di = 0; di < 16; ++di)
			{
				sprintf(hex_output + di * 2, "%02x", digest[di]);
			}

			if(strcmp(cpinfo->verify_md5, hex_output) == 0)
			{
				cpinfo->hash_check = true;

				if(oparg->V == VERBOS)
				{
					printf("%s : %s\n", cpinfo->verify_md5, cpinfo->from);
					printf("%s : %s\n", hex_output, cpinfo->to);
					fflush(stdout);
				}

				if(oparg->L == LOG)
				{
					write_log_hash_check(cpinfo, cpinfo->verify_md5, hex_output, sdir, cpdd, true);
				}
			}
			else
			{
				fprintf(stderr, "ハッシュ値が一致しませんでした、%sを削除します\n", cpinfo->to);
				print_error("NODATA", __FILE__, __LINE__, cpdd);
				write_log_hash_check(cpinfo, cpinfo->verify_md5, hex_output, sdir, cpdd, false);
				cpinfo->write = false;

				errno = 0;
				if(unlink(cpinfo->to) == -1)
				{
					if(errno != ENOENT)
					{
						print_error("unlink", __FILE__, __LINE__, cpdd);
						fprintf(stderr, "%s の削除に失敗しました\n", cpinfo->to);
					}
				}
			}
			break;

		case false:
			fprintf(stderr, "ベリファイ出来ません\n");
			fprintf(stderr, "メモリの確保に失敗しました\n");
			print_error("NODATA", __FILE__, __LINE__, cpdd);
			write_log_hash_check(cpinfo, cpinfo->verify_md5, "NULL", sdir, cpdd, false);
			cpinfo->write = false;
			break;
		}
	}
	else
	{
		fprintf(stderr, "ベリファイに失敗しました\n");
		fprintf(stderr, "%sをオープン出来ません\n", cpinfo->to);
		print_error("NODATA", __FILE__, __LINE__, cpdd);
		write_log_hash_check(cpinfo, cpinfo->verify_md5, "NULL", sdir, cpdd, false);
		cpinfo->write = false;
	}

	if(open_flag == true)
	{
		if(close(fd) == -1)
		{
			print_error("close", __FILE__, __LINE__, cpdd);
		}
	}
/*
	free(buf[0]);
	free(buf[1]);
*/
}

/*******************************************************************************
*******************************************************************************/
static void * read_func(void *dummy)
{
	static int i;
	static int try_count;

	dummy++;	/* コンパイル時のwarningを出さないようにするため */

	i = r_number;
	try_count = 0;

	for(;;)
	{
		read_size[i] = pread(fd, buf[i], buffer_size_local, total);

		if(read_size[i] == 0)
		{
			read_finish = true;
			break;
		}
		else if((read_size[i] % psize) == 0)
		{
			total += read_size[i];

			if(lseek(fd, read_size[i], SEEK_CUR) == -1)
			{
				lseek_err = true;
				break;
			}

			break;
		}
		else if((read_size[i] + total) >= static_to_size)
		{
			total += read_size[i];

			if(lseek(fd, read_size[i], SEEK_CUR) == -1)
			{
				lseek_err = true;
				break;
			}

			break;
		}
		else
		{
			try_count++;

			if(try_count >= 10)
			{
				read_err = true;
				break;
			}
			else
			{
				continue;
			}
		}
	}

	return NULL;
}

/*******************************************************************************
*******************************************************************************/
static void * append_func(void *dummy)
{
	static int i;
	i = a_number;

	md5_append(&state, buf[i], read_size[i]);

	dummy++;	/* コンパイル時のwarningを出さないようにするため */

	return NULL;
}
